unit Prometheus.Client.CollectorRegistry;

interface

uses
  System.Generics.Collections,
  Prometheus.Client.Collector,
  Prometheus.Client.Samples;

type

  TCollectorRegistry = class
  private
    FCollectorsToNames: TDictionary<TCollector, TArray<string>>;
    FNamesToCollectors: TDictionary<string, TCollector>;
    class var FDefaultRegistry: TCollectorRegistry;
    class function GetDefaultRegistry: TCollectorRegistry; static;
  public
    class destructor Finalize;
    constructor Create(AOwnsCollectors: Boolean = True);
    destructor Destroy; override;
    procedure Clear();
    function Collect: TArray<TMetricSamples>;
    function GetCollector<T: TCollector>(const AName: string): T;
    function HasCollector(const AName: string): Boolean;
    procedure &Register(ACollector: TCollector);
    procedure Unregister(ACollector: TCollector);
    class property DefaultRegistry: TCollectorRegistry read GetDefaultRegistry;
  end;

implementation

uses
  System.SysUtils;

function TCollectorRegistry.Collect: TArray<TMetricSamples>;
begin
  SetLength(Result, 0);
  if FCollectorsToNames.Count <= 0 then
    Exit;
  var LSampleList := TList<TArray<TMetricSamples>>.Create;
  try
    for var LCollectorItem in FCollectorsToNames.Keys do
      LSampleList.Add(LCollectorItem.Collect);
    Result := TArray.Concat<TMetricSamples>(LSampleList.ToArray);
  finally
    LSampleList.Free;
  end;
end;

constructor TCollectorRegistry.Create(AOwnsCollectors: Boolean);
begin
  inherited Create;
  if AOwnsCollectors then
    FCollectorsToNames := TObjectDictionary<TCollector, TArray<string>>.Create([doOwnsKeys])
  else
    FCollectorsToNames := TDictionary<TCollector, TArray<string>>.Create;
  FNamesToCollectors := TDictionary<string, TCollector>.Create;
end;

destructor TCollectorRegistry.Destroy;
begin
  if Assigned(FCollectorsToNames) then
    FreeAndNil(FCollectorsToNames);
  if Assigned(FNamesToCollectors) then
    FreeAndNil(FNamesToCollectors);
  inherited Destroy;
end;

class destructor TCollectorRegistry.Finalize;
begin
  if Assigned(FDefaultRegistry) then
    FreeAndNil(FDefaultRegistry);
end;

function TCollectorRegistry.GetCollector<T>(const AName: string): T;
var
  LItem: TCollector;
begin
  if FNamesToCollectors.TryGetValue(AName, LItem) then
    Result := T(LItem)
  else
    Result := nil;
end;

class function TCollectorRegistry.GetDefaultRegistry: TCollectorRegistry;
begin
  if not Assigned(FDefaultRegistry) then
    FDefaultRegistry := TCollectorRegistry.Create();
  Result := FDefaultRegistry;
end;

function TCollectorRegistry.HasCollector(const AName: string): Boolean;
begin
  TMonitor.Enter(Self);
  try
    Result := FNamesToCollectors.ContainsKey(AName);
  finally
    TMonitor.Exit(Self);
  end;
end;

procedure TCollectorRegistry.Clear;
begin
  TMonitor.Enter(Self);
  try
    FCollectorsToNames.Clear;
    FNamesToCollectors.Clear;
  finally
    TMonitor.Exit(Self);
  end;
end;

procedure TCollectorRegistry.Register(ACollector: TCollector);
var
  LName: string;
begin
  TMonitor.Enter(Self);
  try
    var LCollectorNames := ACollector.GetNames;
    for LName in LCollectorNames do
    begin
      if FNamesToCollectors.ContainsKey(LName) then
        raise Exception.Create('Name is already in use by another registered collector');
    end;
    for LName in LCollectorNames do
      FNamesToCollectors.AddOrSetValue(LName, ACollector);
    FCollectorsToNames.AddOrSetValue(ACollector, LCollectorNames);
  finally
    TMonitor.Exit(Self);
  end;
end;

procedure TCollectorRegistry.Unregister(ACollector: TCollector);
begin
  TMonitor.Enter(Self);
  try
    FCollectorsToNames.Remove(ACollector);
    for var LName in ACollector.GetNames do
      FNamesToCollectors.Remove(LName);
  finally
    TMonitor.Exit(Self);
  end;
end;

end.
