unit Prometheus.Client.Collectors.Gauge;

interface

uses
  System.SysUtils,
  Prometheus.Client.SimpleCollector,
  Prometheus.Client.Labels,
  Prometheus.Client.Samples;

type

{ TGaugeChild }

  TGaugeChild = class
  private
    FValue: Double;
  public
    procedure Dec(const AAmount: Double = 1);
    procedure Inc(const AAmount: Double = 1);
    procedure SetDuration(const AProc: TProc);
    procedure SetTo(const AValue: Double);
    procedure SetToCurrentTime;
    property Value: Double read FValue;
  end;

{ TGauge }

  TGauge = class (TSimpleCollector<TGaugeChild>)
  private
    function GetValue: Double;
  protected
    function CreateNewChild: TGaugeChild; override;
  public
    procedure Dec(const AAmount: Double = 1);
    procedure Inc(const AAmount: Double = 1);
    function Collect: TArray<TMetricSamples>; override;
    function GetNames: TArray<string>; override;
    procedure SetTo(const AValue: Double);
    procedure SetToCurrentTime;
    property Value: Double read GetValue;
  end;

implementation

uses
  System.DateUtils,
  System.Diagnostics,
  Prometheus.Client.Resources;

{ TGaugeChild }

procedure TGaugeChild.Dec(const AAmount: Double);
begin
  FValue := FValue - AAmount;
end;

procedure TGaugeChild.Inc(const AAmount: Double);
begin
  FValue := FValue + AAmount;
end;

procedure TGaugeChild.SetDuration(const AProc: TProc);
begin
  if not Assigned(AProc) then
    raise EArgumentNilException.Create(StrErrNullProcReference);
  var LStopWatch := TStopwatch.StartNew;
  try
    AProc;
  finally
    LStopWatch.Stop;
    FValue := LStopWatch.Elapsed.TotalMilliseconds;
  end;
end;

procedure TGaugeChild.SetTo(const AValue: Double);
begin
  FValue := AValue;
end;

procedure TGaugeChild.SetToCurrentTime;
begin
  FValue := TDateTime.NowUTC.ToUnix();
end;

{ TGauge }

function TGauge.Collect: TArray<TMetricSamples>;
begin
  SetLength(Result, 1);
  var LMetric := PMetricSamples(@Result[0]);
  LMetric^.MetricName := Self.Name;
  LMetric^.MetricHelp := Self.Help;
  LMetric^.MetricType := 'gauge';
  SetLength(LMetric^.Samples, ChildrenCount);
  var LIndex := 0;
  EnumChildren(
    procedure (const ALabelValues: TLabelValues; AChild: TGaugeChild)
    begin
      var LSample := PSample(@LMetric^.Samples[LIndex]);
      LSample^.MetricName := Self.Name;
      LSample^.LabelNames := Self.LabelNames;
      LSample^.LabelValues := ALabelValues;
      LSample^.TimeStamp := 0;
      LSample^.Value := AChild.Value;
      System.Inc(LIndex);
    end
  );
end;

function TGauge.CreateNewChild: TGaugeChild;
begin
  Result := TGaugeChild.Create();
end;

procedure TGauge.Dec(const AAmount: Double);
begin
  GetUnlabelledChild.Dec(AAmount);
end;

function TGauge.GetNames: TArray<string>;
begin
  Result := [Name];
end;

function TGauge.GetValue: Double;
begin
  Result := GetUnlabelledChild.Value;
end;

procedure TGauge.Inc(const AAmount: Double);
begin
  GetUnlabelledChild.Inc(AAmount);
end;

procedure TGauge.SetTo(const AValue: Double);
begin
  GetUnlabelledChild.SetTo(AValue);
end;

procedure TGauge.SetToCurrentTime;
begin
  GetUnlabelledChild.SetToCurrentTime;
end;

end.
