unit Prometheus.Client.SimpleCollector;

interface

uses
  System.Generics.Collections,
  System.SysUtils,
  Prometheus.Client.Collector,
  Prometheus.Client.CollectorRegistry,
  Prometheus.Client.Labels,
  Prometheus.Client.Metrics,
  Prometheus.Client.Samples;

type

{ TChildrenCollectProc<TChild> }

  TChildrenCollectProc<TChild: class> = reference to procedure (
    const ALabelValues: TLabelValues; AChild: TChild);

{ TSimpleCollector<TChild> }

  TSimpleCollector<TChild: class> = class abstract(TCollector)
  private
    FChildren: TDictionary<TLabelValues, TChild>;
    FName: string;
    FHelp: string;
    FLabelNames: TLabelNames;
    procedure InitializeNoLabelChildIfNeeded();
    function GetChildrenCount: Integer;
  protected
    function CreateNewChild: TChild; virtual;
    procedure EnumChildren(AProc: TChildrenCollectProc<TChild>);
    function GetUnlabelledChild: TChild;
  public
    constructor Create(const AName: string; const AHelp: string = '';
      const ALabelNames: TLabelNames = []); virtual;
    destructor Destroy; override;
    procedure Clear;
    function Labels(const ALabelValues: TLabelValues): TChild;
    procedure RemoveLabels(const ALabelValues: TLabelValues);
    procedure &Register(ARegistry: TCollectorRegistry = nil);
    procedure Unregister(ARegistry: TCollectorRegistry = nil);
    property Name: string read FName;
    property Help: string read FHelp;
    property ChildrenCount: Integer read GetChildrenCount;
    property LabelNames: TLabelNames read FLabelNames;
  end;

implementation

uses
  Prometheus.Client.Resources;

{ TSimpleCollector<TChild> }

constructor TSimpleCollector<TChild>.Create(const AName: string;
  const AHelp: string = ''; const ALabelNames: TLabelNames = []);
begin
  inherited Create;
  TMetricValidator.CheckName(AName);
  if Length(ALabelNames) > 0 then
    TLabelValidator.CheckLabels(ALabelNames);
  FName := AName;
  FHelp := AHelp;
  FLabelNames := ALabelNames;
  FChildren := TObjectDictionary<TLabelValues, TChild>
    .Create([doOwnsValues], TLabelNamesEqualityComparer.Create); // TODO: Valutare anche passaggio di un comparer per comodit
  InitializeNoLabelChildIfNeeded;
end;

destructor TSimpleCollector<TChild>.Destroy;
begin
  if Assigned(FChildren) then
    FreeAndNil(FChildren);
  inherited Destroy;
end;

function TSimpleCollector<TChild>.CreateNewChild: TChild;
begin
  Result := nil;
end;

procedure TSimpleCollector<TChild>.EnumChildren(AProc: TChildrenCollectProc<TChild>);
begin
  for var LChildPair in FChildren do
    AProc(LChildPair.Key, LChildPair.Value);
end;

function TSimpleCollector<TChild>.GetChildrenCount: Integer;
begin
  Result := FChildren.Count;
end;

function TSimpleCollector<TChild>.GetUnlabelledChild: TChild;
begin
  if Length(FLabelNames) > 0 then
    raise EInvalidOpException.Create('Error Message');
  InitializeNoLabelChildIfNeeded;
  Result := FChildren.Values.ToArray[0];
end;

procedure TSimpleCollector<TChild>.InitializeNoLabelChildIfNeeded;
begin
  if (Length(FLabelNames) <= 0) and (FChildren.Count <= 0) then
    FChildren.Add(nil, CreateNewChild);
end;

function TSimpleCollector<TChild>.Labels(const ALabelValues: TLabelValues): TChild;
begin
  if Length(ALabelValues) <= 0 then
    raise EArgumentException.Create(StrErrLabelValuesMissing);
  if Length(ALabelValues) <> Length(FLabelNames) then
    raise EArgumentException.Create(StrErrLabelNameValueMismatch);
  if FChildren.TryGetValue(ALabelValues, Result) then
    Exit;
  Result := CreateNewChild;
  if not Assigned(Result) then
    Exit;
  try
    FChildren.Add(ALabelValues, Result);
  except
    FreeAndNil(Result);
    raise;
  end;
end;

procedure TSimpleCollector<TChild>.Clear;
begin
  FChildren.Clear;
  InitializeNoLabelChildIfNeeded;
end;

procedure TSimpleCollector<TChild>.Register(ARegistry: TCollectorRegistry);
begin
  if not Assigned(ARegistry) then
    ARegistry := TCollectorRegistry.DefaultRegistry;
  ARegistry.&Register(Self);
end;

procedure TSimpleCollector<TChild>.RemoveLabels(const ALabelValues: TLabelValues);
begin
  FChildren.Remove(ALabelValues);
  InitializeNoLabelChildIfNeeded;
end;

procedure TSimpleCollector<TChild>.Unregister(ARegistry: TCollectorRegistry);
begin
  if not Assigned(ARegistry) then
    ARegistry := TCollectorRegistry.DefaultRegistry;
  ARegistry.Unregister(Self);
end;

end.
