unit Prometheus.Client.Writers.OpenMetrics;

interface

uses
  System.Classes, System.SysUtils,
  Prometheus.Client.Samples;

type

{ TOpenMetricsEncoding }

  TOpenMetricsEncoding = class(TUTF8Encoding)
  public
    function GetPreamble: TBytes; override;
  end;

{ TOpenMetricsWriter }

  TOpenMetricsWriter = class
  private
    function EscapeToken(const AText: string): string;
    function FormatDouble(const AValue: Double): string;
  public
    function Write(AMetricSamples: TArray<TMetricSamples>): string; overload;
    procedure Write(ABuilder: TStringBuilder; AMetricSamples: TArray<TMetricSamples>); overload;
    procedure Write(AWriter: TTextWriter; AMetricSamples: TArray<TMetricSamples>); overload;
    procedure Write(AStream: TStream; AMetricSamples: TArray<TMetricSamples>); overload;
  end;

implementation

uses
  System.Character;

{ TOpenMetricsEncoding }

function TOpenMetricsEncoding.GetPreamble: TBytes;
begin
  SetLength(Result, 0);
end;

{ TOpenMetricsWriter }

function TOpenMetricsWriter.EscapeToken(const AText: string): string;
begin
  Result := AText
    .Replace('\', '\\')
    .Replace(#13#10, '\n')
    .Replace(#13, '\n')
    .Replace(#10, '\n')
    .Replace(#9, '\t')
    .Replace('"', '\"');
end;

function TOpenMetricsWriter.FormatDouble(const AValue: Double): string;
begin
  if AValue.IsNegativeInfinity then
  begin
    Result := '-Inf';
    Exit;
  end;
  if AValue.IsPositiveInfinity then
  begin
    Result := '+Inf';
    Exit;
  end;
  if AValue.IsNan then
  begin
    Result := 'Nan';
    Exit;
  end;
  Result := AValue.ToString;
end;

function TOpenMetricsWriter.Write(AMetricSamples: TArray<TMetricSamples>): string;
begin
  var LBuffer := TStringBuilder.Create;
  try
    Write(LBuffer, AMetricSamples);
    Result := LBuffer.ToString;
  finally
    LBuffer.Free;
  end;
end;

procedure TOpenMetricsWriter.Write(AStream: TStream;
  AMetricSamples: TArray<TMetricSamples>);
begin
  var LEncoding := TOpenMetricsEncoding.Create();
  try
    var LStreamWriter := TStreamWriter.Create(AStream, LEncoding);
    try
      Write(LStreamWriter, AMetricSamples);
    finally
      LStreamWriter.Free;
    end;
  finally
    LEncoding.Free;
  end;
end;

procedure TOpenMetricsWriter.Write(ABuilder: TStringBuilder;
  AMetricSamples: TArray<TMetricSamples>);
begin
  var LStringWriter := TStringWriter.Create(ABuilder);
  try
    Write(LStringWriter, AMetricSamples);
  finally
    LStringWriter.Free;
  end;
end;

procedure TOpenMetricsWriter.Write(AWriter: TTextWriter;
  AMetricSamples: TArray<TMetricSamples>);
begin
  for var LMFS in AMetricSamples do
  begin
    AWriter.Write('# HELP');
    AWriter.Write(' ');
    AWriter.Write(LMFS.MetricName);
    AWriter.Write(' ');
    AWriter.Write(EscapeToken(LMFS.MetricHelp));
    if not LMFS.MetricHelp.EndsWith('.') then
      AWriter.Write('.');
    AWriter.Write(#10);

    AWriter.Write('# TYPE');
    AWriter.Write(' ');
    AWriter.Write(LMFS.MetricName);
    AWriter.Write(' ');
    AWriter.Write(LMFS.MetricType);
    AWriter.Write(#10);

    for var LSample in LMFS.Samples do
    begin
      AWriter.Write(LSample.MetricName);
      if LSample.HasLabels then
      begin
        AWriter.Write('{');
        for var LIndex := 0 to Pred(Length(LSample.LabelNames)) do
        begin
          if LIndex > 0 then
            AWriter.Write(',');
          AWriter.Write(LSample.LabelNames[LIndex]);
          AWriter.Write('="');
          AWriter.Write(EscapeToken(LSample.LabelValues[LIndex]));
          AWriter.Write('"');
        end;
        AWriter.Write('}');
      end;
      AWriter.Write(' ');
      AWriter.Write(FormatDouble(LSample.Value));
      if LSample.HasTimeStamp then
      begin
        AWriter.Write(' ');
        AWriter.Write(LSample.TimeStamp);
      end;
      AWriter.Write(#10);
    end;
  end;
end;

end.
